/*
 *  libtriplex.c: implementation section of the triplex search
 *             program. Searches for potential triplex sequences in DNA.
 *             can be used as a library or together with triplex.c
 *             to compile into a command-line application.
 *  Copyright (C) 2011
 *    Brno University of Technology, Faculty of Information Technology
 *    Masaryk University, Faculty of Informatics
 *
 *  Author(s): Tomas Martinek <martinto@fit.vutbr.cz>
 *             Daniel Kopecek <xkopecek@fi.muni.cz>
 *             Matej Lexa <lexa@fi.muni.cz> 
 *             
 *  LICENSE TERMS
 *               
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *  3. Neither the name of the Company nor the names of its contributors
 *  may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *                                              
 *  This software is provided ``as is'', and any express or implied
 *  warranties, including, but not limited to, the implied warranties of
 *  merchantability and fitness for a particular purpose are disclaimed.
 *  In no event shall the company or contributors be liable for any
 *  direct, indirect, incidental, special, exemplary, or consequential
 *  damages (including, but not limited to, procurement of substitute
 *  goods or services; loss of use, data, or profits; or business
 *  interruption) however caused and on any theory of liability, whether
 *  in contract, strict liability, or tort (including negligence or
 *  otherwise) arising in any way out of the use of this software, even
 *  if advised of the possibility of such damage.
 */

#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include <stdbool.h>
#include <inttypes.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "assume.h"
#include "sm_alloc.h"
#include "libtriplex.h"
#include "gff3.h"

/* status variable flags */
#define STAT_NONE         0
#define STAT_QUALITY      1
#define STAT_MINLEN       2
#define STAT_EXPORT       4

/* Penalization scores */
/* DP rule constants */
#define DP_MATCH        0
#define DP_MISMATCH     1
#define DP_LEFT         2
#define DP_RIGHT        3

/* Penalization scores */
#define PEN_BOUND       3
#define PEN_MISMATCH    3
#define PEN_INSDEL      5

#define DTWIST_LIMIT    7

#define NUM_TRI_TYPES   8

/* --------------------------------------------------------------------------
 *                          Type definition
 * --------------------------------------------------------------------------
 */

/* DP matrix position represented as diagonal and antidiagonal */
typedef struct{
        int     diag;
        int     antidiag;
} t_pos;

/* Data stored for every diagonal */
typedef struct{
        t_pos   start;  /* Position of the first match */
        t_pos   max_score_pos;  /* Maximal score position */
        int     score;  /* Actual score */
        int     max_score;  /* Maximal score */
        int     status; /* Status */
        int     bound;  /* Type of the triplex bound */
        int     twist;  /* angle between C1 atoms */
        int     dtwist;  /* Change in angle between subsequent triplets */
        int     dp_rule;  /* Previous position: 0 - match, 1 - mismatch, 2 - left, 3 - right  */
        int     indels;  /* number of indels */
        int     max_indels;  /* number of indels from start position to max score postition */
} t_diag;

/*
Tritype
 0: parallel triplex on sense strand
 1: parallel triplex on antisense strand
 2: antiparallel triplex on sense strand
 3: antiparallel triplex on antisense strand
 4: parallel triplex on sense strand
 5: parallel triplex on antisense strand
 6: antiparallel triplex on sense strand
 7: antiparallel triplex on antisense strand
*/

/* Lookup tables */
const int tab_score[NUM_TRI_TYPES][4][4] = {
           /* A  C  G  T */
 /* T */   { {2, 1, 1, -9},
 /* G */     {-9, -9, 1, 2},
 /* C */     {-9, -9, 2, -9},
 /* A */     {-9, -9, -9, -9} },

           /* T  G  C  A */
 /* A */   { {-9, -9, -9, -9},
 /* C */     {-9, 2, -9, -9},
 /* G */     {2, 1, -9, -9},
 /* T */     {-9, 1, 1, 2} },

           /* A  C  G  T */
 /* A */   { {2, -9, 1, -9},
 /* C */     {1, -9, -9, -9},
 /* G */     {-9, -9, 2, -9},
 /* T */     {2, 1, -9, -9} },

           /* T  G  C  A */
 /* T */   { {-9, -9, 1, 2},
 /* G */     {-9, 2, -9, -9},
 /* C */     {-9, -9, -9, 1},
 /* A */     {-9, 1, -9, 2} },

           /* A  C  G  T */
 /* T */   { {2, -9, -9, -9},
 /* G */     {1, -9, -9, -9},
 /* C */     {1, 1, 2, -9},
 /* A */     {-9, 2, -9, -9} },

           /* T  G  C  A */
 /* A */   { {-9, -9, 2, -9},
 /* C */     {-9, 2, 1, 1},
 /* G */     {-9, -9, -9, 1},
 /* T */     {-9, -9, -9, 2} },

           /* A  C  G  T */
 /* A */   { {2, 1, -9, 2},
 /* C */     {-9, -9, -9, 1},
 /* G */     {1, -9, 2, -9},
 /* T */     {-9, -9, -9, -9} },

           /* T  G  C  A */
 /* T */   { {-9, -9, -9, -9},
 /* G */     {-9, 2, -9, 1},
 /* C */     {1, -9, -9, -9},
 /* A */     {2, -9, 1, 2} },
};

const int tab_bound[NUM_TRI_TYPES][4][4] = {
           /* A  C  G  T */
 /* T */   { {1, 2, 1, 0},
 /* G */     {0, 0, 2, 2},
 /* C */     {0, 0, 1, 0},
 /* A */     {0, 0, 0, 0} },

           /* T  G  C  A */
 /* A */   { {0, 0, 0, 0},
 /* C */     {0, 1, 0, 0},
 /* G */     {2, 2, 0, 0},
 /* T */     {0, 1, 2, 1} },

           /* A  C  G  T */
 /* A */   { {1, 0, 2, 0},
 /* C */     {2, 0, 0, 0},
 /* G */     {0, 0, 3, 0},
 /* T */     {1, 3, 0, 0} },

           /* T  G  C  A */
 /* T */   { {0, 0, 3, 1},
 /* G */     {0, 3, 0, 0},
 /* C */     {0, 0, 0, 2},
 /* A */     {0, 2, 0, 1} },

           /* A  C  G  T */
 /* T */   { {1, 0, 0, 0},
 /* G */     {2, 0, 0, 0},
 /* C */     {1, 2, 1, 0},
 /* A */     {0, 2, 0, 0} },

           /* T  G  C  A */
 /* A */   { {0, 0, 2, 0},
 /* C */     {0, 1, 2, 1},
 /* G */     {0, 0, 0, 2},
 /* T */     {0, 0, 0, 1} },

           /* A  C  G  T */
 /* A */   { {1, 2, 0, 1},
 /* C */     {0, 0, 0, 3},
 /* G */     {2, 0, 3, 0},
 /* T */     {0, 0, 0, 0} },

           /* T  G  C  A */
 /* T */   { {0, 0, 0, 0},
 /* G */     {0, 3, 0, 2},
 /* C */     {3, 0, 0, 0},
 /* A */     {1, 0, 2, 1} },
};

const int tab_twist[NUM_TRI_TYPES][4][4] = {
           /* A  C  G  T */
 /* T */   { {71, 26, 83, 33},
 /* G */     {99, 56, 93, 45},
 /* C */     {34, 24, 49, 23},
 /* A */     {71, 31, 59, 53} },

           /* T  G  C  A */
 /* A */   { {53, 59, 31, 71},
 /* C */     {23, 49, 24, 34},
 /* G */     {45, 93, 56, 99},
 /* T */     {33, 83, 26, 71} },

           /* A  C  G  T */
 /* A */   { {81, 64, 118, 59},
 /* C */     {126, 80, 126, 70},
 /* G */     {4, 43, 75, 40},
 /* T */     {103, 45, 116, 51} },

           /* T  G  C  A */
 /* T */   { {51, 116, 45, 103},
 /* G */     {40, 75, 43, 4},
 /* C */     {70, 126, 80, 126},
 /* A */     {59, 118, 64, 81} },

           /* A  C  G  T */
 /* T */   { {71, 99, 34, 71},
 /* G */     {26, 56, 24, 31},
 /* C */     {83, 93, 49, 59},
 /* A */     {33, 45, 23, 53} },

           /* T  G  C  A */
 /* A */   { {53, 23, 45, 33},
 /* C */     {59, 49, 93, 83},
 /* G */     {31, 24, 56, 26},
 /* T */     {71, 34, 99, 71} },

           /* A  C  G  T */
 /* A */   { {81, 126, 4, 103},
 /* C */     {64, 80, 43, 45},
 /* G */     {118, 126, 75, 116},
 /* T */     {59, 70, 40, 51} },

           /* T  G  C  A */
 /* T */   { {51, 40, 70, 59},
 /* G */     {116, 75, 126, 118},
 /* C */     {45, 43, 80, 64},
 /* A */     {103, 4, 126, 81} },
};

/* Default values */
// double LAMBDA[NUM_TRI_TYPES] = {0.67, 0.67, 0.71, 0.71, 0.67, 0.67, 0.71, 0.71};
//double MI[NUM_TRI_TYPES]     = {6.05, 6.05, 5.88, 5.88, 6.05, 6.05, 5.88, 5.88}; 

/* E. coli values */
/* double LAMBDA[NUM_TRI_TYPES] = {0.65, 0.65, 0.74, 0.74, 0.65, 0.65, 0.74, 0.74}; */
/* double MI[NUM_TRI_TYPES]     = {6.09, 6.09, 5.78, 5.78, 6.09, 6.09, 5.78, 5.78}; */

/* Human */
double LAMBDA[NUM_TRI_TYPES] = {0.68, 0.68, 0.68, 0.68, 0.68, 0.68, 0.68, 0.68};
double MI[NUM_TRI_TYPES]     = {5.98, 5.98, 5.97, 5.97, 5.98, 5.98, 5.97, 5.97};

/* --------------------------------------------------------------------------
 *                          Procedures and function 
 * -------------------------------------------------------------------------- 
 */
double p_function(int score, int tri_type) {
        return 1-exp(-exp(-LAMBDA[tri_type]*(score-MI[tri_type])));
}

double e_value(int score, int tri_type) {
        return p_function(score, tri_type);
}

double p_value(int score, int tri_type) {
        return 1-exp(-p_function(score, tri_type));
}

/**
 * Calcutates the length of a path in DP matrix
 */
static int get_length(int start_antidiag, int end_antidiag, int start_diag, int end_diag)
{
        int length;

        length = (end_antidiag-start_antidiag+2)/2;
        /* diagonal odd to even - increment length */
        if (((start_diag % 2) == 1) && ((end_diag % 2) == 0))
                length++;
        if (length == 0)
                length = 1;

        return length;
}

/**
 * Application of DP rule
 */
static t_diag get_max_score(char a, char b, t_diag dl, t_diag d, t_diag dr, int diag, int antidiag, int tri_type, int max_loop)
{
        t_diag res;
        int incscore, mm_score;

        /* --------------------------------------------------------------- */
        /* match or mismatch score computation  */
        incscore = tab_score[tri_type][a][b]; 
          // a trick to have zero or negative scores 
          // for low quality triplets, changed all incscore comparisons to -9
        if (incscore> -9) {  /* match */
                mm_score = d.score+incscore;
//                if ((tab_bound[tri_type][a][b] != d.bound) && (d.dp_rule == DP_MATCH))
                if ((tab_bound[tri_type][a][b] != d.bound) && (abs(tab_twist[tri_type][a][b] - d.twist) > DTWIST_LIMIT ) && (abs(tab_twist[tri_type][a][b] - d.twist + d.dtwist) > DTWIST_LIMIT)  && (d.dp_rule == DP_MATCH))
                        mm_score -= PEN_BOUND;
        } else { /* mismatch */
                mm_score = d.score-PEN_MISMATCH;
        }

        /* match/mismatch is better */
        if ((mm_score >= dl.score-PEN_INSDEL) && (mm_score >= dr.score-PEN_INSDEL)) {
                res = d;
                res.score = mm_score;
                res.dp_rule = DP_MISMATCH;

                /* match */
                if (incscore>-9) {
                        res.dp_rule = DP_MATCH;
                        res.bound = tab_bound[tri_type][a][b];
                        res.twist = tab_twist[tri_type][a][b];
                        res.dtwist = res.twist - d.twist;
                        if (mm_score >= res.max_score) {
                                res.max_score = mm_score;
                                res.max_score_pos.diag = diag;
                                res.max_score_pos.antidiag = antidiag;
                                res.max_indels = res.indels;
                        }
                }
        }
        /* --------------------------------------------------------------- */
        /* insetion or deletion */
        else {
                /* get from left diagonal */
                if (dl.score > dr.score) {
                        res = dl;
                        res.score = dl.score-PEN_INSDEL;
                        res.dp_rule = DP_LEFT;
                }
                /* get from right diagonal */
                else {
                        res = dr;
                        res.score = dr.score-PEN_INSDEL;
                        res.dp_rule = DP_RIGHT;
                }
                res.indels++;

        }
        /* --------------------------------------------------------------- */
        /* local alignment - only for loop */
        if ((res.score < 0) && (antidiag <= max_loop)) {
                res.score = 0;
                res.max_score = 0;
                res.start.diag = diag;
                res.start.antidiag = antidiag;
                res.max_score_pos.diag = diag;
                res.max_score_pos.antidiag = antidiag;
                res.indels = 0;
                res.max_indels = 0;
        }

        /* --------------------------------------------------------------- */
        return res;
}

static void export_data_gff3(gff3_t *gff, const char *seq_id, t_diag *diag, bool exported, char *data, int tri_type)
{
        gff3_rec_t *rec;
	uint64_t gap_start, gap_end;
	int length;
	char numbuf[32+1];

        rec = gff3_rec_new();

        rec->seq_id = strdup(seq_id);
        rec->source = strdup("triplex");
        rec->type   = strdup("match");

        rec->end   = (uint64_t)((diag->max_score_pos.diag + diag->max_score_pos.antidiag - 1)/2);
        rec->start = (uint64_t)(rec->end - diag->max_score_pos.antidiag);

	++rec->end;
	++rec->start;

	gap_end   = (diag->start.diag + diag->start.antidiag - 1)/2;
	gap_start = gap_end - diag->start.antidiag;

	++gap_end;
	++gap_start;

	length    = get_length(diag->start.antidiag, diag->max_score_pos.antidiag,
			       diag->start.diag,     diag->max_score_pos.diag);

        rec->score  = (double)diag->max_score;
        rec->strand = '+';
        rec->phase  = -1;

	gff3_rec_attr_add(rec, "exported", exported ? "1" : "0");

	snprintf(numbuf, sizeof numbuf - 1, "%"PRIu64, gap_start);
	gff3_rec_attr_add(rec, "gap_start", strdup(numbuf));

	snprintf(numbuf, sizeof numbuf - 1, "%"PRIu64, gap_end);
	gff3_rec_attr_add(rec, "gap_end",   strdup(numbuf));

	snprintf(numbuf, sizeof numbuf - 1, "%d", length);
	gff3_rec_attr_add(rec, "length",    strdup(numbuf));

	snprintf(numbuf, sizeof numbuf - 1, "%.8g", p_value(diag->max_score, tri_type));
	gff3_rec_attr_add(rec, "p_val",     strdup(numbuf));

	snprintf(numbuf, sizeof numbuf - 1, "%d", diag->indels);
	gff3_rec_attr_add(rec, "indels",    strdup(numbuf));

	/* gff3_rec_attr_add(rec, "seq", palindrome); */

        gff3_add_raw(gff, rec);

        sm_free(rec);

        return;
}

int triplex_search(gff3_t *gff, const char *buf_id, char *buffer, size_t buflen, tsearch_params_t *params)
{
        register size_t i, d, x;
	register int c, length, num;
	int exp_num, nexp_num;

	t_diag *diag;

	assume_d(gff    != NULL, -1);
	assume_d(buffer != NULL, -1);
	assume_d(params != NULL, -1);

	/*
	 * Allocate and initialize the diag array
	 */
	diag = sm_alloc(2 * buflen * sizeof(t_diag));

	if (diag == NULL)
		return (-1);

	for (i = 0;  i < 2 * buflen; ++i) {
                diag[i].score     = 0;
                diag[i].max_score = 0;
                diag[i].bound     = 0;
                diag[i].twist     = 90;
                diag[i].dtwist    = 0;
                diag[i].status    = STAT_NONE;
                diag[i].start.diag     = i;
                diag[i].start.antidiag = (((params->min_loop+i) % 2) == 0) ? params->min_loop+1 : params->min_loop+2;
                diag[i].max_score_pos.diag     = diag[i].start.diag;
                diag[i].max_score_pos.antidiag = diag[i].start.antidiag;
                diag[i].indels     = 0;
                diag[i].max_indels = 0;
                diag[i].dp_rule    = DP_MISMATCH;

                /* minloop    diag       antidiag */
                /* sude       sude         +0     */
                /* sude       liche        +1 */
                /* liche      sude         +1 */
                /* liche      liche        +0 */
	}

	/*
	 * Main computation loop
	 */
	exp_num = 0;

	for (x = params->min_loop + 1; x <= (size_t)(2 * params->steps) && x < buflen; ++x) {
		d = x + 1;

		for (i=x; i < buflen; i++) {
			/*
			 * max score and length calcualtion
			 */
			diag[d] = get_max_score(buffer[i], buffer[i - x],
						diag[d - 1], diag[d], diag[d + 1],
						d, x, params->triplex_type, params->max_loop);

                        length = get_length(diag[d].start.antidiag, diag[d].max_score_pos.antidiag,
                                            diag[d].start.diag,     diag[d].max_score_pos.diag);

			diag[d].status = (length >= params->min_length) ?
				diag[d].status | STAT_MINLEN : diag[d].status & (~STAT_MINLEN);

			/*
			 * actual score satisfies the required quality
			 */
			if (diag[d].score >= params->min_score) {
				diag[d].status |= STAT_QUALITY;

				/*
				 * if triplex can not continue, then export
				 */
				if ((diag[d].status & STAT_MINLEN) &&
				    (d == (x+1) || d == (2 * buflen- x - 1)))
				{
					exp_num++;
					diag[d].status = STAT_EXPORT;

					if (p_value(diag[d].max_score, params->triplex_type) < params->p_val)
						export_data_gff3(gff, buf_id, diag + d, true, buffer, params->triplex_type);
				}
			} else {/* actual score does not satisfy the required quality */

				/*
				 * if quality requirement was satisfied
				 * in previous step, then export
				 */
				if((!(diag[d-1].status & STAT_QUALITY)) &&
				   (!(diag[d+1].status & STAT_QUALITY)) &&
				   ( (diag[ d ].status & STAT_QUALITY)) &&
				   ( (diag[ d ].status & STAT_MINLEN)))
				{
					exp_num++;
					diag[d].status = STAT_EXPORT;

					if (p_value(diag[d].max_score, params->triplex_type) < params->p_val)
						export_data_gff3(gff, buf_id, diag + d, true, buffer, params->triplex_type);

					diag[d].max_score = 0;
				} else {
					diag[d].status = STAT_NONE;
				}
			}

			d += 2;
		}
	}

        nexp_num = 0;

        for (i = 1; (size_t)i < (2 * buflen); i++) {
                if ((diag[i].status & STAT_QUALITY) &&
                    (diag[i].status & STAT_MINLEN))
                {
                        if(p_value(diag[i].max_score, params->triplex_type) < params->p_val)
                                export_data_gff3(gff, buf_id, diag + i, false, buffer, params->triplex_type);

                        nexp_num++;
                }
        }

	sm_free(diag);

	return (0);
}
